(* week-13_from-concrete-data-types-to-abstract-data-types-revisited.ml *)
(* Introduction to Computer Science (YSC1212), Sem1, 2020-2021 *)
(* Olivier Danvy <danvy@yale-nus.edu.sg> *)
(* Version of Sat 03 Apr 2021 *)

(* ********** *)

module type INT_LIST =
  sig
    type int_list
    val nil : int_list
    val is_empty : int_list -> bool
    val equal : int_list -> int_list -> bool
    val cons : int -> int_list -> int_list
    val car : int_list -> int option
    val cdr : int_list -> int_list option
  end;;

let test_int_list nil is_empty equal cons car cdr =
  let b0 = (is_empty nil = true)
  and b1 = (is_empty (cons 2 nil) = false)
  and b2 = (equal nil nil = true)
  and b3 = (equal (cons 1 nil) nil = false)
  and b4 = (equal nil (cons 2 nil) = false)
  and b5 = (equal (cons 1 nil) (cons 2 nil) = false)
  and b6 = (equal (cons 1 nil) (cons 1 nil) = true)
  and b7 = (equal (cons 2 (cons 1 nil)) (cons 2 (cons 1 nil)) = true)
  and b8 = (car nil = None)
  and b9 = (car (cons 1 nil) = Some 1)
  and b10 = (match car (cons 1 nil) with
             | None ->
                false
             | Some n ->
                n = 1)
  and b11 = (match cdr (cons 2 (cons 1 nil)) with
             | None ->
                false
             | Some ns ->
                equal ns (cons 1 nil))
  (* etc. *)
  in b0 && b1 && b2 && b3 && b4 && b5 && b6 && b7 && b8 && b9 && b10 && b11;;

(* ********** *)

module Int_List_native : INT_LIST =
  struct
    type int_list = int list

    let nil = []

    let is_empty is =
      is = []

    let equal is js =
      is = js

    let cons i is =
      i :: is

    let car is =
      match is with
      | [] ->
         None
      | i :: _ ->
         Some i

    let cdr is =
      match is with
      | [] ->
         None
      | _ :: is' ->
         Some is'
  end;;

let () = assert (test_int_list Int_List_native.nil
                               Int_List_native.is_empty
                               Int_List_native.equal
                               Int_List_native.cons
                               Int_List_native.car
                               Int_List_native.cdr);;

(* ********** *)

module Int_List_l2r : INT_LIST =
  struct
    type int_list =
      | Nil
      | Cons of int * int_list

    let nil = Nil

    let is_empty is =
      is = Nil

    let equal is js =
      is = js

    let cons i is =
      Cons (i, is)

    let car is =
      match is with
      | Nil ->
         None
      | Cons (i, _) ->
         Some i

    let cdr is =
      match is with
      | Nil ->
         None
      | Cons (_, is') ->
         Some is'
  end;;

let () = assert (test_int_list Int_List_l2r.nil
                               Int_List_l2r.is_empty
                               Int_List_l2r.equal
                               Int_List_l2r.cons
                               Int_List_l2r.car
                               Int_List_l2r.cdr);;

(* ********** *)

module Int_List_r2l : INT_LIST =
  struct
    type int_list =
      | Nil
      | Cons of int_list * int

    let nil = Nil

    let is_empty is =
      is = Nil

    let equal is js =
      is = js

    let cons i is =
      Cons (is, i)

    let car is =
      match is with
      | Nil ->
         None
      | Cons (_, i) ->
         Some i

    let cdr is =
      match is with
      | Nil ->
         None
      | Cons (is', _) ->
         Some is'
  end;;

let () = assert (test_int_list Int_List_r2l.nil
                               Int_List_r2l.is_empty
                               Int_List_r2l.equal
                               Int_List_r2l.cons
                               Int_List_r2l.car
                               Int_List_r2l.cdr);;

(* ********** *)

(* end of week-13_from-concrete-data-types-to-abstract-data-types-revisited.ml *)

let end_of_file = "week-13_from-concrete-data-types-to-abstract-data-types-revisited.ml";;

(*
           OCaml version 4.01.0
   
   # #use "week-13_from-concrete-data-types-to-abstract-data-types-revisited.ml";;
   module type INT_LIST =
     sig
       type int_list
       val nil : int_list
       val is_empty : int_list -> bool
       val equal : int_list -> int_list -> bool
       val cons : int -> int_list -> int_list
       val car : int_list -> int option
       val cdr : int_list -> int_list option
     end
   val test_int_list : 'a -> ('a -> bool) -> ('a -> 'a -> bool) -> (int -> 'a -> 'a) -> ('a -> int option) -> ('a -> 'a option) -> bool = <fun>
   module Int_List_native : INT_LIST
   module Int_List_l2r : INT_LIST
   module Int_List_r2l : INT_LIST
   val end_of_file : string = "week-13_from-concrete-data-types-to-abstract-data-types-revisited.ml"
   # 
*)
